/*******************************
 * Main file for the pressure sensitive alarmclock.
 * Author: Cooper Bills <cbills07@gmail.com>
 * July 2010
 *******************************/
 
 //************************
 // Software Configuration:
 
int8_t timezone = -5; //time offset from GMT time 
byte useDST = 1; //does your timezone use daylight savings time?
#define GRACE_PERIOD 1200000 //im ms, time after getting up that bed stays "hot"
#define BEEP_LENGTH 500 //in ms, length of a beep on the alarm
#define PRESSURE_INITAL_OFFSET 20 //value to offset the threshold when reset

 //*****************
 // Hardware Pinout:
 
 // Pressure sensor (higher voltage with more pressure) to analog pin: 
byte sensor1Pin = 0;
  
 // Speaker or buzzer connected to digital pin:
byte speakerPin = 11; // for a speaker that requires a PWM
byte buzzerPin = 8; // for a buzzer that only requires a high input to make noise

 // I2C Real time clock module (RTC)
#include <Wire.h> //included here since the Arduino IDE requires it here
#include "rtc.h" //uses the Wire library:
                //analog pin 4 for SDA; and analog pin 5 for SCL

 // WWVB signal interrupts
#include "wwvb.h" //uses interrupt0 (digital pin 2)
                  //This library also configures Timer1
byte wwvbStatusLED = 13; //this pin will go high with a high interupt, and low with a low interupt
    //a solid looking pulse means good signal.
                  
 // LCD
#include <LiquidCrystal.h>
byte lcdPins[] = {4, 5, 6, 7}; //DB4-DB7 digital pin numbers
byte lcdE = 9; //LCD operation enable pin
byte lcdRS = 10; //LCD Register select signal
              //R/W pin of LCD should just be grounded
              
 // Buttons
#include "button.h"
byte incPin = 3; //button for incrementing currently selected value
byte togPin = 12; //button for toggling currently selected value
              
 // End Configuration/Pinout              
 /*******************************/
 
#include <pins_arduino.h>
#include "pressureSensor.h"
#include <EEPROM.h>

#define ALARM_HOURS_ADDRESS 0 // Address in EEPROM to load/store alarm state
#define ALARM_MINUTES_ADDRESS 1
#define ALARM_ENABLED_ADDRESS 2
byte alarm_minutes, alarm_hours, alarm_enabled, alarm_changed, alarm_activated;
unsigned long alarm_grace_target, alarm_last_beep_toggle;

LiquidCrystal lcd(lcdRS, lcdE, lcdPins[0], lcdPins[1], lcdPins[2], lcdPins[3]);
PressureSensor sensor1(sensor1Pin, PRESSURE_INITAL_OFFSET);
Button incButton(incPin);
Button moveButton(togPin);

char lcdBuffer[9];
unsigned long lastLCDUpdate = 0;
unsigned int syncs = 0;

#define NUMOFSCREENS 4 //how many different screens we will have
byte currentScreen = 0; //0 = time; 1 = alarm; 2 = status
byte currentScreenPos = 0; //currently selected item on screen
byte screenPosCount[] = {5, 5, 1, 1}; //options per screen

void synched_callback(unsigned char sec, unsigned char min, unsigned char hrs, 
                      unsigned int day, unsigned char year, unsigned char dst)
{
  syncs++;
  setRTCTime(sec, min, hrs + timezone + (useDST ? dst : 0));
}

// The setup() method runs once, when the sketch starts
void setup()   {    
  // initalize the PWM
  // using Arduino PWMs, but we want a higher frequency, so we'll raise the prescaler:
  //TCCR2B |= (1<<CS21) | (1<<CS20); // set prescaler to clk/256 (lower frequency)
  //TCCR2B &= ~((1<<CS22));
  //TCCR2B |= (1<<CS21); // set prescaler to clk/128
  //TCCR2B &= ~((1<<CS22) | (1<<CS20));
  //ASSR &= ~(1<<AS2); // Use internal clock - external clock not used in Arduino
  
  // Get Alarm Settings from EEPROM:
  alarm_hours = EEPROM.read(ALARM_HOURS_ADDRESS);
  alarm_minutes = EEPROM.read(ALARM_MINUTES_ADDRESS);
  alarm_enabled = EEPROM.read(ALARM_ENABLED_ADDRESS);
  alarm_changed = 0;
  alarm_activated = 0;
  pinMode(buzzerPin, OUTPUT);
  
  Serial.begin(9600);  
  lcd.begin(8, 2);
  setup_wwvb(synched_callback, wwvbStatusLED);
  setup_rtc();
  sensor1.resetThreshold(PRESSURE_INITAL_OFFSET);
  sensor1.setDropBelowThesholdCallback(offBed);
  incButton.setPressedCallback(incButton_callback);
  moveButton.setPressedCallback(moveButton_callback);
  Serial.println("initialized");
}

void loop()
{  
  sensor1.update();
  incButton.update();
  moveButton.update();
  updateAlarm();
  updateSound();
  loop_wwvb();
  loop_rtc();
  updateLCD(0);
}

void offBed()
{
  alarm_grace_target = millis() + GRACE_PERIOD;
}

void updateAlarm()
{
  if(alarm_enabled)
  {
    if(getRTCMinutes() == alarm_minutes && getRTCHours() == alarm_hours)
      alarm_activated = 1;
      
    if(alarm_activated && !sensor1.isAboveThreshold() && abs(alarm_grace_target - millis()) < 1000)
    {
      sensor1.resetThreshold(PRESSURE_INITAL_OFFSET);
      alarm_activated = 0;
    }
  }
  else
    alarm_activated = 0;
}

int alarmVolume = 1;

void updateSound()
{
  if(alarm_activated && sensor1.isAboveThreshold())
  {
    if(abs(alarm_last_beep_toggle - millis()) > BEEP_LENGTH/2)
    {
      alarm_last_beep_toggle = millis() + BEEP_LENGTH/2;
      //***Start Buzzer Code
      if(alarmVolume <= 0) digitalWrite(buzzerPin, LOW); //disable the buzzer when alarm is disabled
      else if(alarmVolume > 0 && alarmVolume <= 61) //occasional pulses
      {
        if(!((alarmVolume-1) % 20)) digitalWrite(buzzerPin, HIGH);
        else digitalWrite(buzzerPin, LOW);
      }
      else if(alarmVolume > 61 && alarmVolume <= 93) //more occasional pulses
      {
        if(!((alarmVolume-61) % 8)) digitalWrite(buzzerPin, HIGH);
        else digitalWrite(buzzerPin, LOW);
      }
      else if(alarmVolume > 93 && alarmVolume < 127) //more frequent pulses
      {
        if(!((alarmVolume-93) % 6)) digitalWrite(buzzerPin, HIGH);
        else digitalWrite(buzzerPin, LOW);
      }
      else //frequent pulses
      {
        if(TCNT1 % 30) //add some randomness to the pattern, to be extra annoying (psudo-random, about 1/30 chance of not toggling)
          *portInputRegister(digitalPinToPort(buzzerPin)) = digitalPinToBitMask(buzzerPin); //toggle buzzer each tick
      }
      //***End Buzzer Code
      
      if(alarmVolume< 127) alarmVolume += 2; //if alarmVolume is currently active and not at max, increase it.
      
      //***Start Speaker Code
      if(PORTB ^ 0x30)
        analogWrite(speakerPin, 0); 
      else
        analogWrite(speakerPin, alarmVolume); //uses Arduino's PWMs
      //***End Speaker Code
    }
  }
  else
  {
    alarmVolume = 1;
    analogWrite(speakerPin, 0);
    digitalWrite(buzzerPin, LOW);
  }
}
    

void updateLCD(byte newData)
{
  if(newData || abs(millis() - lastLCDUpdate) >= 500)
  {
    lastLCDUpdate = millis();
    
    lcd.setCursor(0, 0);
    if(currentScreen == 0) //time screen
      sprintf(lcdBuffer, "%2d:%02d:%02d", getRTCHours(), getRTCMinutes(), getRTCSeconds());
    else if(currentScreen == 1)
      sprintf(lcdBuffer, "A: %2d:%02d", alarm_hours, alarm_minutes);
    else if(currentScreen == 2)
      sprintf(lcdBuffer, "T|Syncs:");
    else if(currentScreen == 3)
      sprintf(lcdBuffer, "Val:%4d", sensor1.getValue());
    lcd.print(lcdBuffer);
    Serial.println(lcdBuffer);
    
    lcd.setCursor(0, 1);
    if(currentScreen == 0) //time screen
      for(int i = 0; i < 8; i++)
        if(i == currentScreenPos + (currentScreenPos / 2) || (currentScreenPos == 4 && i == 7))
          lcdBuffer[i] = '^';
        else
          lcdBuffer[i] = ' ';
    else if(currentScreen == 1)
    {
      if(alarm_enabled) sprintf(lcdBuffer, "On      ");
      else sprintf(lcdBuffer, "Off     ");
      if(currentScreenPos == 0) lcdBuffer[3] = '<';
      else lcdBuffer[(currentScreenPos + 2 + (currentScreenPos/3))%8] = '^';
    }
    else if(currentScreen == 2)
      sprintf(lcdBuffer, "%1d|%06d", sensor1.isAboveThreshold(), syncs);
    else if(currentScreen == 3)
      sprintf(lcdBuffer, "Thr:%4d", sensor1.getThreshold());
    lcd.print(lcdBuffer);
    Serial.println(lcdBuffer);
   
    if(!newData && (millis()/500)%2 && getRTCVL()) //if on odd half of second, and VL is flagged (time not set)
    {
      lcd.clear(); //clear the lcd (flashing effect)
      Serial.println("---");
    }
  }
}
  

void moveButton_callback()
{
  if(alarm_changed)
  {
    EEPROM.write(ALARM_HOURS_ADDRESS, alarm_hours);
    EEPROM.write(ALARM_MINUTES_ADDRESS, alarm_minutes);
  }
  
  currentScreenPos++;
  if(currentScreenPos >= screenPosCount[currentScreen])
  {
    currentScreenPos = 0;
    if(getRTCVL())  //if time is not set, lock to time screen
      currentScreen = 0;
    else
      currentScreen = (currentScreen + 1) % NUMOFSCREENS;
  }
  updateLCD(1);
}

void incButton_callback()
{
  switch(currentScreen)
  {
    case 0: //Time Screen
      {
        byte hours = getRTCHours();
        byte minutes = getRTCMinutes();
        byte seconds = getRTCSeconds();
        switch(currentScreenPos)
        {
          case 0:  //Time Hours 10's
            hours += 10;
            if(hours >= 24)
              hours = hours % 10;
            break;
          case 1:  //Time Hours 1's
            hours++;
            if(hours >= 24)
              hours = 20;
            else if(!(hours % 10))
              hours -= 10;
            break;
          case 2:  //Time Minutes 10's
            minutes += 10;
            if(minutes >= 60)
              minutes = minutes % 10;
            break;
          case 3: //Time Minutes 1's
            minutes++;
            if(!(minutes % 10))
              minutes -= 10;
            break;
          case 4: //Time Seconds
            seconds = 0;
            break;
        }
        setRTCTime(seconds, minutes, hours);
        updateLCD(1);
      }
      break; //end Time Screen
    case 1: //Alarm Screen
      switch(currentScreenPos)
      {
        case 0: //alarm on/off
          alarm_enabled = !alarm_enabled;
          EEPROM.write(ALARM_ENABLED_ADDRESS, alarm_enabled);
          break;
        case 1:  //Alarm Hours 10's
          alarm_hours += 10;
          if(alarm_hours >= 24)
            alarm_hours = alarm_hours % 10;
          alarm_changed = 1;
          break;
        case 2:  //Alarm Hours 1's
          alarm_hours++;
          if(alarm_hours >= 24)
            alarm_hours = 20;
          else if(!(alarm_hours % 10))
            alarm_hours -= 10;
          alarm_changed = 1;
          break;
        case 3:  //Alarm Minutes 10's
          alarm_minutes += 10;
          if(alarm_minutes >= 60)
            alarm_minutes = alarm_minutes % 10;
          alarm_changed = 1;
          break;
        case 4: //Alarm Minutes 1's
          alarm_minutes++;
          if(!(alarm_minutes % 10))
            alarm_minutes -= 10;
          alarm_changed = 1;
          break;
      }
      updateLCD(1);
      break; //end alarm screen
    case 2: //status Screen
      //Nothing to do here.
      updateLCD(1);
      break;
    case 3: //threshold screen
      sensor1.resetThreshold(PRESSURE_INITAL_OFFSET);
      updateLCD(1);
      break;
  }
}
