#ifndef _WWVB_H_
#define _WWVB_H_
/*******************************
 * Functions/Library for receiving a WWVB signal.
 * Author: Cooper Bills <cbills07@gmail.com>
 * July 2010
 *
 * A callback function matching this spec: */
typedef void(* clockSyncCallback) (unsigned char sec, unsigned char min, unsigned char hrs, 
                                  unsigned int day, unsigned char year, unsigned char dst);
/* Should be passed to init_wwvb().  The callback will get called when 
 * A successful packet of time data arrives, and will be passed
 * the exact time at the moment it's called.
 *******************************/

enum WWVBBIT{
  ZERO,
  ONE,
  SYNC,
  JUNK,
  OVER
};

volatile WWVBBIT wwvb_timedata[60];
volatile WWVBBIT lastWWVBBIT = JUNK;
volatile WWVBBIT lastWWVBBITBuffer = JUNK;
volatile unsigned char wwvb_newSync = 0;
volatile unsigned char wwvb_lastinc = 0;
volatile unsigned char wwvb_bitinc;
volatile int wwvb_timeMark = -1; //position in sequence
volatile unsigned char wwvb_newbit = 1;
volatile unsigned char wwvb_led = 0;

volatile clockSyncCallback wwvbCallback = 0;

//prototypes
void wwvbFlip();
void wwvbRise();
void processBit(int inWWVBBIT);
void wwvbFall();
void doWWVBCallback();

void setup_wwvb(clockSyncCallback callback, unsigned char statusLED)
{
  wwvb_led = statusLED;
  if(wwvb_led)
    pinMode(wwvb_led, OUTPUT);
  wwvbCallback = callback;
  
  // initalize the Time Timer (timer1) (this is old code.)
  //TCCR1A &= ~((1<<WGM13) | (1<<WGM12) | (1<<WGM11) | (1<<WGM10));  //Normal Operation (no pwm)
  //TCCR1B |= (1<<CS12) | (1<<CS10); //set clk/1024 prescaling
  //TCCR1B &= ~(1<<CS11); //set clk/1024 prescaling
  //ASSR &= ~(1<<AS2); // Use internal clock - external clock not used in Arduino
  
  for(int i = 0; i < 60; i++)
    wwvb_timedata[i] = JUNK;
    
  attachInterrupt(0, wwvbFlip, CHANGE);
}

void loop_wwvb()
{
  if(wwvb_lastinc != wwvb_bitinc)
  {
    wwvb_lastinc = wwvb_bitinc;
    processBit(lastWWVBBITBuffer);
  }
  if(wwvb_newSync)
  {
    wwvb_newSync = 0;
    doWWVBCallback();
  }
}
  
void doWWVBCallback()
{
  //if any bits are junk, then signal is unstable, so we will not do the callback
  int i;
  for(i = 0; i < 59; i++)
    if(wwvb_timedata[i] == JUNK || wwvb_timedata[i] == OVER)
      return;
      
  unsigned char sec = 59;
  unsigned char min = 0;
  unsigned char hour = 0;
  unsigned int day = 0;
  unsigned char year = 0;
  unsigned char dst = 0;
  
  if(wwvb_timedata[1] == ONE) min += 40;
  if(wwvb_timedata[2] == ONE) min += 20;
  if(wwvb_timedata[3] == ONE) min += 10;
  if(wwvb_timedata[5] == ONE) min += 8;
  if(wwvb_timedata[6] == ONE) min += 4;
  if(wwvb_timedata[7] == ONE) min += 2;
  if(wwvb_timedata[8] == ONE) min += 1;
 
  if(wwvb_timedata[12] == ONE) hour += 20;
  if(wwvb_timedata[13] == ONE) hour += 10;
  if(wwvb_timedata[15] == ONE) hour += 8;
  if(wwvb_timedata[16] == ONE) hour += 4;
  if(wwvb_timedata[17] == ONE) hour += 2;
  if(wwvb_timedata[18] == ONE) hour += 1;
  
  if(wwvb_timedata[22] == ONE) day += 200;
  if(wwvb_timedata[23] == ONE) day += 100;
  if(wwvb_timedata[25] == ONE) day += 80;
  if(wwvb_timedata[26] == ONE) day += 40;
  if(wwvb_timedata[27] == ONE) day += 20;
  if(wwvb_timedata[28] == ONE) day += 10;
  if(wwvb_timedata[30] == ONE) day += 8;
  if(wwvb_timedata[31] == ONE) day += 4;
  if(wwvb_timedata[32] == ONE) day += 2;
  if(wwvb_timedata[33] == ONE) day += 1;
  
  if(wwvb_timedata[45] == ONE) year += 80;
  if(wwvb_timedata[46] == ONE) year += 40;
  if(wwvb_timedata[47] == ONE) year += 20;
  if(wwvb_timedata[48] == ONE) year += 10;
  if(wwvb_timedata[50] == ONE) year += 8;
  if(wwvb_timedata[51] == ONE) year += 4;
  if(wwvb_timedata[52] == ONE) year += 2;
  if(wwvb_timedata[53] == ONE) year += 1;
  
  if(wwvb_timedata[57] == ONE) dst += 1;
  
  wwvbCallback(sec, min, hour, day, year, dst);
}

void processBit(int inWWVBBIT)
{
  WWVBBIT in = (WWVBBIT)inWWVBBIT;
  /*
  switch(in)
  {
    case JUNK:
      Serial.println("Recived WWVB Junk");  
      break;
    case OVER:
      Serial.println("Recived WWVB Over");  
      break;
    case ONE:
      Serial.println("Recived WWVB One");  
      break;
    case ZERO:
      Serial.println("Recived WWVB Zero");  
      break;
    case SYNC:
      Serial.println("Recived WWVB Sync");  
      break;
  }
  */
  if(wwvb_timeMark < 0) //out of sync
  {
    if(in == SYNC)  //and new bit is SYNC
      wwvb_timeMark = 0; //begin processing stream
    return;
  }   
  
  if((wwvb_timeMark % 10) == 9 || !wwvb_timeMark) //if it should be a sync bit
  {
    if(in != SYNC)
    {
      wwvb_timeMark = -1; //out of sync
      return;
    }
  }
  wwvb_timedata[wwvb_timeMark] = in;
  wwvb_timeMark++;
  if(wwvb_timeMark >= 59) 
  {
    doWWVBCallback();
    wwvb_timeMark = -1;
  }
}
  
void wwvbFlip()
{
  char val = digitalRead(2);
  if(wwvb_led)
    digitalWrite(wwvb_led, val);
  if(val == HIGH)
    wwvbRise();
  else
    wwvbFall();
}

volatile unsigned int riseTime = 0;
volatile unsigned int fallTime = 0;
volatile unsigned char lastSuccess = 2;

//fall should happen on the second
void wwvbFall()
{
  if(lastWWVBBIT == JUNK)
    return;
  if(lastWWVBBIT == OVER)
  {
    fallTime = millis();
    return;  
  }
  
  unsigned int ticks = millis() - riseTime;
  unsigned int lowTicks, highTicks;
  
  if(lastWWVBBIT == ZERO)
  {
    lowTicks = 780;
    highTicks = 840;
  }
  else if(lastWWVBBIT == ONE)
  {
    lowTicks = 485;
    highTicks = 535;
  }
  else //sync
  {
    lowTicks = 190;
    highTicks = 230;
  }
  
  if(ticks < lowTicks)
    ;//Serial.println("Junk in falling code"); //its junk, wait for next falling edge
  else if(ticks >= lowTicks && ticks <= highTicks)
    {fallTime = millis(); lastWWVBBITBuffer = lastWWVBBIT; wwvb_bitinc++; wwvb_newbit = 1;} //proper time has occured, so validate (inc wwvb_bitinc)
  else
    {/*Serial.println("fallLate");*/ fallTime = millis(); lastWWVBBIT = OVER; lastWWVBBITBuffer = lastWWVBBIT; wwvb_bitinc++; wwvb_newbit = 1;}//Serial.println("f");} //we went overtime >.<
}

void wwvbRise()
{
  if(lastSuccess == wwvb_bitinc) //waiting for good fall time
    return;
  
  if(wwvb_newbit)
    {riseTime = millis(); wwvb_newbit = 0;}
  
  unsigned int ticks = millis() - fallTime;
  
  if(ticks < 185)
    lastWWVBBIT = JUNK;
  else if(ticks < 210)
    {lastWWVBBIT = ZERO; lastSuccess = wwvb_bitinc;}
  else if(ticks < 475)
    lastWWVBBIT = JUNK;
  else if(ticks < 520)
    {lastWWVBBIT = ONE; lastSuccess = wwvb_bitinc;}
  else if(ticks < 765)
    lastWWVBBIT = JUNK;
  else if(ticks < 830)
    {lastWWVBBIT = SYNC; lastSuccess = wwvb_bitinc;}
  else 
  {
    //Serial.println("riseLate");
    lastWWVBBIT = OVER;
    lastWWVBBITBuffer = OVER;
    lastSuccess = wwvb_bitinc;
    wwvb_bitinc++;
    wwvb_newbit = 1;
  }
  
  //if(lastWWVBBIT == JUNK)
  //  Serial.println("Junk found in rise code");
}

#endif
